import 'dart:io';

import 'package:dio/dio.dart';
import 'package:dio_cache_interceptor/dio_cache_interceptor.dart';
import 'package:flutter_rest_client/flutter_rest_client.dart';

import 'dio_builder.dart';

abstract class IHttpHelper {
  Future<dynamic> request(IRequestEndPoint endPoint, IRequestModel requestModel,
      {Map<String, dynamic> headers,
      bool cacheRequest = true,
      String? cacheSubKey,
      bool tryRefreshToken = true,
      Map<String, dynamic> queriesParams});
}

class HttpHelper implements IHttpHelper {
  final Dio _dio;
  CancelToken cancelToken = CancelToken();
  IHttpConfig config;
  final options = CacheOptions(
    store: MemCacheStore(),
    policy: CachePolicy.refreshForceCache,
    hitCacheOnErrorExcept: [401, 403],
    maxStale: const Duration(days: 7),
    priority: CachePriority.normal,
    cipher: null,
    keyBuilder: CacheOptions.defaultCacheKeyBuilder,
    allowPostMethod: false,
  );

  HttpHelper({required this.config})
      : _dio = DioBuilder(config: config).build();

  @override
  Future request(IRequestEndPoint endPoint, IRequestModel requestModel,
      {Map<String, dynamic>? headers,
      bool cacheRequest = true,
      String? cacheSubKey,
      Map<String, dynamic>? queriesParams,
      bool tryRefreshToken = true}) async {
    try {
      switch (endPoint.method) {
        case RequestMethod.DELETE:
          return (await _dio.delete(endPoint.url,
                  options: Options(
                    headers: headers,
                    followRedirects: false,
                    validateStatus: (status) {
                      return status! < 500;
                    },
                  ),
                  queryParameters: queriesParams,
                  data: requestModel.toJson(),
                  cancelToken: cancelToken))
              .data;
        case RequestMethod.GET:
          if (cacheRequest) {
            _dio..interceptors.add(DioCacheInterceptor(options: options));
          }
          return (await _dio.get(endPoint.url,
                  options: cacheRequest
                      ? options
                          .copyWith(
                            policy: CachePolicy.refreshForceCache,
                          )
                          .toOptions()

                      // buildConfigurableCacheOptions(
                      //         options: Options(headers: headers,followRedirects: false,
                      //           validateStatus: (status) {
                      //             return status! < 500;
                      //           },),
                      //         subKey: cacheSubKey,
                      //         forceRefresh: true)
                      : Options(
                          headers: headers,
                          followRedirects: false,
                          validateStatus: (status) {
                            return status! < 500;
                          },
                        ),
                  // queryParameters: requestModel.toJson(),
                  queryParameters: queriesParams,
                  cancelToken: cancelToken))
              .data;

        case RequestMethod.POST:
          return (await _dio.post(endPoint.url,
                  options: Options(
                    headers: headers,
                    followRedirects: false,
                    validateStatus: (status) {
                      return status! < 500;
                    },
                  ),
                  data: requestModel.toJson(),
                  queryParameters: queriesParams,
                  cancelToken: cancelToken))
              .data;
        case RequestMethod.PATCH:
          return (await _dio.patch(endPoint.url,
                  options: Options(
                    headers: headers,
                    followRedirects: false,
                    validateStatus: (status) {
                      return status! < 500;
                    },
                  ),
                  data: requestModel.toJson(),
                  queryParameters: queriesParams,
                  cancelToken: cancelToken))
              .data;
        case RequestMethod.PUT:
          return (await _dio.put(endPoint.url,
                  options: Options(
                    headers: headers,
                    followRedirects: false,
                    validateStatus: (status) {
                      return status! < 500;
                    },
                  ),
                  data: requestModel.toJson(),
                  queryParameters: queriesParams,
                  cancelToken: cancelToken))
              .data;
      }
    } on DioError catch (e) {
      switch (e.response?.statusCode) {
        case 400:
        case 401:
        case 403:
          config.listener.clearSession();
      }
      if (e.response?.statusCode == 401 && tryRefreshToken == true) {
        print('session expired trying refresh token');
        final refreshToken = await config.refreshToken;
        final accessToken = await config.token;
        if (refreshToken != null && accessToken != null) {
          final refreshed =
              await refreshAccessToken(config.refreshTokenUrl, refreshToken);
          if (refreshed == true)
            return request(endPoint, requestModel,
                cacheRequest: cacheRequest, tryRefreshToken: false);
        } else {
          config.listener.clearSession();
        }
      }
      return e.response?.data;
    } on SocketException catch (e) {
      throw SocketException(e.toString());
    } on FormatException catch (_) {
      throw FormatException();
    } catch (e) {
      throw e;
    }
  }

  Future<bool?> refreshAccessToken(String url, String refreshToken) async {
    try {
      final response =
          await _dio.post(url, data: {'refreshToken': refreshToken});
      if (response.statusCode == 200 && response.data['accessToken'] != null) {
        config.onTokenRefreshed(response.data['accessToken']);
        return true;
      }
    } catch (e) {}
    return false;
  }

  ///auth/refresh-token
}
